from typing import Any, Dict

import numpy as np
import regional


def spots_to_geojson(spots_viz):
    '''
    Convert spot geometrical data to geojson format
    '''

    def make_dict(row):
        row = row[1]
        d = dict()
        d['properties'] = {'id': int(row.spot_id), 'radius': int(row.r)}
        d['geometry'] = {'type': 'Point', 'coordinates': [int(row.x), int(row.y)]}
        return d

    return [make_dict(row) for row in spots_viz.iterrows()]


def geojson_to_region(geojson: Dict[Any, Any]) -> regional.many:
    """
    Convert geojson data to region geometrical data.
    """
    def make_region(geometry):
        assert geometry['geometry']['type'] == "Polygon"
        region = [(coordinates[0], coordinates[1])
                  for coordinates in geometry['geometry']['coordinates']]

        return regional.one(region)

    return regional.many([make_region(geometry) for geometry in geojson])


def relabel(image):
    '''
    This is a local implementation of centrosome.cpmorphology.relabel
    to remove this dependency from starfish.

    Takes a labelled image and relabels each image object consecutively.
    Original code from:
    https://github.com/CellProfiler/centrosome/blob/master/centrosome/cpmorphology.py

    They use a BSD-3 license, which would then have to be propagated to starfish,
    this could be an issue.

    Args
    ----
    image: numpy.ndarray
        A 2d integer array representation of an image with labels

    Returns
    -------
    new_image: numpy.ndarray
        A 2d integer array representation of an image wiht new labels

    n_labels: int
        The number of new unique labels
    '''

    # I've set this as a separate function, rather than binding it to the
    # _WatershedSegmenter object for now

    unique_labels = set(image[image != 0])
    n_labels = len(unique_labels)

    # if the image is unlabelled, return original image
    # warning/message required?
    if n_labels == 0:
        return (image, 0)

    consec_labels = np.arange(n_labels) + 1
    lab_table = np.zeros(max(unique_labels) + 1, int)
    lab_table[[x for x in unique_labels]] = consec_labels

    # Use the label table to remap all of the labels
    new_image = lab_table[image]

    return new_image, n_labels



from io import BytesIO
from typing import IO, Tuple

import numpy as np
from skimage.io import imsave
from slicedimage import (
    ImageFormat,
)


class FetchedTile:
    """
    This is the contract for providing the data for constructing a :class:`slicedimage.Tile`.
    """
    @property
    def shape(self) -> Tuple[int, ...]:
        raise NotImplementedError()

    @property
    def format(self) -> ImageFormat:
        raise NotImplementedError()

    @property
    def tile_data_handle(self) -> IO:
        raise NotImplementedError()


class TileFetcher:
    """
    This is the contract for providing the image data for constructing a
    :class:`slicedimage.Collection`.
    """
    def get_tile(self, fov: int, hyb: int, ch: int, z: int) -> FetchedTile:
        """
        Given fov, hyb, ch, and z, return an instance of a :class:`.FetchedImage` that can be
        queried for the image data.
        """
        raise NotImplementedError()


class RandomNoiseTile(FetchedTile):
    """
    This is a simple implementation of :class:`.FetchedImage` that simply regenerates random data
    for the image.
    """
    @property
    def shape(self) -> Tuple[int, ...]:
        return 1536, 1024

    @property
    def format(self) -> ImageFormat:
        return ImageFormat.TIFF

    @property
    def tile_data_handle(self) -> IO:
        arr = np.random.randint(0, 256, size=self.shape, dtype=np.uint8)
        output = BytesIO()
        imsave(output, arr, plugin="tifffile")
        output.seek(0)
        return output


class RandomNoiseTileFetcher(TileFetcher):
    """
    This is a simple implementation of :class:`.ImageFetcher` that simply returns a
    :class:`.RandomNoiseImage` for every fov, hyb, ch, z combination.
    """
    def get_tile(self, fov: int, hyb: int, ch: int, z: int) -> FetchedTile:
        return RandomNoiseTile()



import sys

from validate_sptx.validate_sptx import validate


class Cli:
    parser_group = None

    @staticmethod
    def add_to_parser(parser):
        """adds experiment-json and fuzz arguments to the given parser"""
        parser.add_argument(
            "--experiment-json",
            required=True,
            metavar="JSON_FILE_OR_URL")
        parser.add_argument(
            "--fuzz", action="store_true")
        parser.set_defaults(starfish_command=Cli.run)
        Cli.parser_group = parser

    @staticmethod
    def run(args, print_help=False):
        """invokes validate with the parsed commandline arguments"""
        try:
            valid = validate(args.experiment_json, fuzz=args.fuzz)
            if valid:
                sys.exit(0)
            else:
                sys.exit(1)
        except KeyboardInterrupt:
            sys.exit(3)



import json
import os
from typing import Tuple

from slicedimage import (
    Collection,
    ImageFormat,
    Tile,
    TileSet,
    Writer,
)

from starfish.experiment import Experiment
from starfish.types import Coordinates, Indices
from .imagedata import FetchedTile, RandomNoiseTile, RandomNoiseTileFetcher, TileFetcher


AUX_IMAGE_NAMES = {
    'nuclei',
    'dots',
}


def tile_opener(toc_path, tile, file_ext):
    tile_basename = os.path.splitext(toc_path)[0]
    return open(
        "{}-Z{}-H{}-C{}.{}".format(
            tile_basename,
            tile.indices[Indices.Z],
            tile.indices[Indices.ROUND],
            tile.indices[Indices.CH],
            ImageFormat.TIFF.file_ext,
        ),
        "wb")


def tile_writer(tile, fh):
    tile.copy(fh)
    return ImageFormat.TIFF


def fov_path_generator(parent_toc_path, toc_name):
    toc_basename = os.path.splitext(os.path.basename(parent_toc_path))[0]
    return os.path.join(
        os.path.dirname(parent_toc_path),
        "{}-{}.json".format(toc_basename, toc_name),
    )


def build_image(
        fov_count, hyb_count, ch_count, z_count,
        image_fetcher: TileFetcher,
        default_shape: Tuple[int, int]
):
    """
    Build and returns an image set with the following characteristics:

    Parameters
    ----------
    fov_count : int
        Number of fields of view in this image set.
    hyb_count : int
        Number for hybridizations in this image set.
    ch_count : int
        Number for channels in this image set.
    z_count : int
        Number of z-layers in this image set.
    default_shape : Tuple[int, int]
        Default shape of the individual tiles in this image set.

    Returns
    -------
    The slicedimage collection representing the image.
    """
    collection = Collection()
    for fov_ix in range(fov_count):
        fov_images = TileSet(
            [Indices.X, Indices.Y, Indices.Z, Indices.ROUND, Indices.CH],
            {Indices.ROUND: hyb_count, Indices.CH: ch_count, Indices.Z: z_count},
            default_shape,
            ImageFormat.TIFF,
        )

        for z_ix in range(z_count):
            for hyb_ix in range(hyb_count):
                for ch_ix in range(ch_count):
                    image = image_fetcher.get_tile(fov_ix, hyb_ix, ch_ix, z_ix)
                    tile = Tile(
                        {
                            Coordinates.X: (0.0, 0.0001),
                            Coordinates.Y: (0.0, 0.0001),
                            Coordinates.Z: (0.0, 0.0001),
                        },
                        {
                            Indices.Z: z_ix,
                            Indices.ROUND: hyb_ix,
                            Indices.CH: ch_ix,
                        },
                        image.shape,
                    )
                    tile.set_source_fh_contextmanager(image.tile_data_handle, image.format)
                    fov_images.add_tile(tile)
        collection.add_partition("fov_{:03}".format(fov_ix), fov_images)
    return collection


def write_experiment_json(
        path,
        fov_count,
        hyb_dimensions,
        aux_name_to_dimensions,
        primary_tile_fetcher=None,
        aux_tile_fetcher=None,
        postprocess_func=None,
        default_shape=(1536, 1024),
):
    """
    Build and returns a top-level experiment description with the following characteristics:

    Parameters
    ----------
    path : str
        Directory to write the files to.
    fov_count : int
        Number of fields of view in this experiment.
    hyb_dimensions : Mapping[str, int]
        Dictionary mapping dimension name to dimension size for the hybridization image.
    aux_name_to_dimensions : Mapping[str, Mapping[str, int]]
        Dictionary mapping the auxiliary image type to dictionaries, which map from dimension name
        to dimension size.
    primary_tile_fetcher : Optional[ImageFetcher]
        ImageFetcher for hybridization images.  Set this if you want specific image data to be set
        for the hybridization images.  If not provided, the image data is set to random noise via
        :class:`RandomNoiseImageFetcher`.
    aux_tile_fetcher : Optional[Mapping[str, ImageFetcher]]
        ImageFetchers for auxiliary images.  Set this if you want specific image data to be set for
        one or more aux image types.  If not provided for any given aux image, the image data is
        set to random noise via :class:`RandomNoiseImageFetcher`.
    postprocess_func : Optional[Callable[[dict], dict]]
        If provided, this is called with the experiment document for any postprocessing.
        An example of this would be to add something to one of the top-level extras field.
        The callable should return what is to be written as the experiment document.
    default_shape : Tuple[int, int] (default = (1536, 1024))
        Default shape for the tiles in this experiment.
    """
    if primary_tile_fetcher is None:
        primary_tile_fetcher = RandomNoiseTileFetcher()
    if aux_tile_fetcher is None:
        aux_tile_fetcher = {}
    if postprocess_func is None:
        postprocess_func = lambda doc: doc

    experiment_doc = {
        'version': str(Experiment.CURRENT_VERSION),
        'auxiliary_images': {},
        'extras': {},
    }
    hybridization_image = build_image(
        fov_count,
        hyb_dimensions[Indices.ROUND], hyb_dimensions[Indices.CH], hyb_dimensions[Indices.Z],
        primary_tile_fetcher,
        default_shape=default_shape,
    )
    Writer.write_to_path(
        hybridization_image,
        os.path.join(path, "hybridization.json"),
        pretty=True,
        partition_path_generator=fov_path_generator,
        tile_opener=tile_opener,
        tile_writer=tile_writer,
    )
    experiment_doc['primary_images'] = "hybridization.json"

    for aux_name, aux_dimensions in aux_name_to_dimensions.items():
        if aux_dimensions is None:
            continue
        auxiliary_image = build_image(
            fov_count,
            aux_dimensions[Indices.ROUND], aux_dimensions[Indices.CH], aux_dimensions[Indices.Z],
            aux_tile_fetcher.get(aux_name, RandomNoiseTileFetcher()),
            default_shape=default_shape,
        )
        Writer.write_to_path(
            auxiliary_image,
            os.path.join(path, "{}.json".format(aux_name)),
            pretty=True,
            partition_path_generator=fov_path_generator,
            tile_opener=tile_opener,
            tile_writer=tile_writer,
        )
        experiment_doc['auxiliary_images'][aux_name] = "{}.json".format(aux_name)

    experiment_doc = postprocess_func(experiment_doc)
    with open(os.path.join(path, "experiment.json"), "w") as fh:
        json.dump(experiment_doc, fh, indent=4)



import pytest
from pkg_resources import resource_filename

from .util import SpaceTxValidator

experiment_schema_path = resource_filename("validate_sptx", "schema/experiment.json")
validator = SpaceTxValidator(experiment_schema_path)
example = resource_filename("validate_sptx", "examples/experiment/experiment.json")


def test_fov():
    assert validator.validate_file(example)


def test_nuclei_must_be_present():
    wrong_nuclei = validator.load_json(example)
    wrong_nuclei['auxiliary_images'] = {'not_nuclei': 'nuclei.json'}
    with pytest.warns(UserWarning):
        assert not validator.validate_object(wrong_nuclei)


def test_version_must_be_semantic():
    wrong_version = validator.load_json(example)
    wrong_version['version'] = '10a'
    with pytest.warns(UserWarning):
        assert not validator.validate_object(wrong_version)


def test_dartfish_example_experiment():
    dartfish_example = resource_filename(
        "validate_sptx", "examples/experiment/dartfish_experiment.json")
    assert validator.validate_file(dartfish_example)



"""
Tests for codebook.from_json method
"""

import json
import os
import tempfile
from typing import Any, Dict, List

from starfish import Codebook
from starfish.types import Features, Indices


def codebook_json_data_factory() -> List[Dict[str, Any]]:
    """
    Codebook with two codewords describing an experiment with three channels and two imaging rounds.
    Both codes have two "on" channels.
    """
    codebook_data = [
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 0, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_A"
        },
        {
            Features.CODEWORD: [
                {Indices.ROUND.value: 0, Indices.CH.value: 2, Features.CODE_VALUE: 1},
                {Indices.ROUND.value: 1, Indices.CH.value: 1, Features.CODE_VALUE: 1}
            ],
            Features.TARGET: "GENE_B"
        },
    ]
    return codebook_data


def test_codebook_loads_from_local_file() -> None:
    """
    dumps the codebook data to a temporary json file and reads it back into a Codebook,
    verifying that the data has not changed.
    """

    # dump codebook to disk
    codebook_data: List = codebook_json_data_factory()
    with tempfile.TemporaryDirectory() as directory:
        codebook_json: str = os.path.join(directory, 'simple_codebook.json')
        with open(codebook_json, 'w') as f:
            json.dump(codebook_data, f)

        # load the codebook
        codebook = Codebook.from_json(codebook_json)
        assert codebook.sizes[Indices.ROUND] == 2
        assert codebook.sizes[Indices.CH] == 3
        assert codebook.sizes[Features.TARGET] == 2


def test_codebook_serialization():
    """
    Test that codebook can be saved to disk and recovered, and that the recovered codebook is
    identical to the one that it was serialized from.
    """
    # Create a codebook
    codebook_array = codebook_json_data_factory()
    codebook = Codebook.from_code_array(codebook_array)

    # Dump it to a temporary file
    with tempfile.TemporaryDirectory() as directory:
        json_codebook = os.path.join(directory, 'codebook.json')
        codebook.to_json(json_codebook)

        # Retrieve it and test that the data it contains has not changed
        codebook_reloaded = Codebook.from_json(json_codebook)
        assert codebook_reloaded.equals(codebook)



import os
import shutil
import subprocess
import tempfile
from typing import Callable, Sequence, Union

from starfish.util import clock


def stages(commands: Sequence[Sequence[Union[str, Callable]]],
           subdirs: Sequence[str]=None,

           keep_data: bool=False) -> str:
    """
    Execute a list of commands in a temporary directory
    cleaning them up unless otherwise requested.

    Parameters
    ----------
    commands : Sequence[Sequence[Union[str, Callable]]]
        A collection of tuples of commands composed either of
        str elements or callable objects which will be invoked
        with the keyword argument "tempdir". The resulting list
        of strings will be passed to subprocess.check_call.
    subdirs : Sequence[str]
        A collection of paths which should be created as subdirectories
        within the temporary directory used by this invocation.
    keep_data : bool
        If not true, shutil.rmtree will be called on the temporary
        directory used by this invocation.

    Return
    ------
    str :
        Path to the temporary directory used by this invocation.

    Environment variables
    ---------------------
    STARFISH_COVERAGE:
         If set, then command lists will have `coverage run ...` options
         prepended before execution.
    """
    tempdir = tempfile.mkdtemp()

    def callback(interval):
        print(" ".join(stage[:2]), " ==> {} seconds".format(interval))

    try:

        if subdirs:
            for subdir in subdirs:
                os.makedirs("{tempdir}".format(
                    tempdir=os.path.join(tempdir, subdir)))

        for stage in commands:
            cmdline = prepare_stage(stage, tempdir)
            with clock.timeit(callback):
                subprocess.check_call(cmdline)

        return tempdir

    finally:
        if not keep_data:
            shutil.rmtree(tempdir)


def prepare_stage(stage: Sequence[Union[str, Callable]],
                  tempdir: str) -> Sequence[str]:
    """
    Loop through elements of stage, building them into a commandline.
    If an element is a callable, it will be invoked with the "tempdir"
    keyword.

    Parameters
    ----------
    stage: Sequence[Union[str, Callable]]
        A collection of commands composed either of
        str elements or callable objects which will be invoked
        with the keyword argument "tempdir". The resulting list
        of strings will be passed to subprocess.check_call.
    tempdir: str
        Temporary directory that will be used by the invoking method.

    Return
    ------
    Sequence of strings for passing to subprocess.check_call

    Environment variables
    ---------------------
    STARFISH_COVERAGE:
         If set, then command lists will have `coverage run ...` options
         prepended before execution.
    """
    coverage_enabled = "STARFISH_COVERAGE" in os.environ
    cmdline = [
        element(tempdir=tempdir) if callable(element) else element
        for element in stage
    ]
    if cmdline[0] == "starfish" and coverage_enabled:
        coverage_cmdline = [
            "coverage", "run",
            "-p",
            "--source", "starfish",
            "-m", "starfish",
        ]
        coverage_cmdline.extend(cmdline[1:])
        cmdline = coverage_cmdline
    elif cmdline[0] == "validate-sptx" and coverage_enabled:
        coverage_cmdline = [
            "coverage", "run",
            "-p",
            "--source", "validate_sptx",
            "-m", "validate_sptx",
        ]
        coverage_cmdline.extend(cmdline[1:])
        cmdline = coverage_cmdline
    return cmdline



from copy import deepcopy

import numpy as np

from starfish.stack import ImageStack
from starfish.util.synthesize import SyntheticData


def divide(array, value):
    return array / value


def test_apply():
    """test that apply correctly applies a simple function across 2d tiles of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert (stack.numpy_array == 1).all()
    stack.apply(divide, value=2)
    assert (stack.numpy_array == 0.5).all()


def test_apply_3d():
    """test that apply correctly applies a simple function across 3d volumes of a Stack"""
    stack = ImageStack.synthetic_stack()
    assert np.all(stack.numpy_array == 1)
    stack.apply(divide, value=4, is_volume=True)
    assert (stack.numpy_array == 0.25).all()


def test_apply_labeled_dataset():
    """
    test that apply correctly applies a simple function across starfish-generated synthetic data
    """
    original = SyntheticData().spots()
    image = deepcopy(original)
    image.apply(divide, value=2)
    assert np.all(image.numpy_array == original.numpy_array / 2)


def test_apply_not_in_place():
    """
    test that apply correctly applies a simple function across a starfish stack without modifying
    original data
    """
    image = SyntheticData().spots()
    new = image.apply(divide, value=2, in_place=False)
    assert np.all(new.numpy_array == image.numpy_array / 2)



import argparse
from functools import partial
from typing import Callable, Optional, Tuple, Union

import numpy as np
from skimage.filters import gaussian

from starfish.stack import ImageStack
from starfish.types import Number
from ._base import FilterAlgorithmBase
from .util import preserve_float_range, validate_and_broadcast_kernel_size


class GaussianLowPass(FilterAlgorithmBase):

    def __init__(
            self, sigma: Union[Number, Tuple[Number]], is_volume: bool=False, **kwargs) -> None:
        """Multi-dimensional low-pass gaussian filter.

        Parameters
        ----------
        sigma : Union[Number, Tuple[Number]]
            Standard deviation for Gaussian kernel.
        is_volume : bool
            If True, 3d (z, y, x) volumes will be filtered, otherwise, filter 2d tiles
            independently.

        """
        self.sigma = validate_and_broadcast_kernel_size(sigma, is_volume)
        self.is_volume = is_volume

    @classmethod
    def add_arguments(cls, group_parser: argparse.ArgumentParser) -> None:
        group_parser.add_argument(
            "--sigma", type=float, help="standard deviation of gaussian kernel")
        group_parser.add_argument(
            "--is-volume", action="store_true",
            help="indicates that the image stack should be filtered in 3d")

    @staticmethod
    def low_pass(image: np.ndarray, sigma: Union[Number, Tuple[Number]]) -> np.ndarray:
        """
        Apply a Gaussian blur operation over a multi-dimensional image.

        Parameters
        ----------
        image : np.ndarray[np.uint16]
            2-d or 3-d image data
        sigma : Union[Number, Tuple[Number]]
            Standard deviation of the Gaussian kernel that will be applied. If a float, an
            isotropic kernel will be assumed, otherwise the dimensions of the kernel give (z, y, x)

        Returns
        -------
        np.ndarray :
            Blurred data in same shape as input image, converted to np.uint16 dtype.

        """

        filtered = gaussian(
            image,
            sigma=sigma, output=None, cval=0, multichannel=False, preserve_range=True, truncate=4.0
        )

        filtered = preserve_float_range(filtered)

        return filtered

    def run(
            self, stack: ImageStack, in_place: bool=True, verbose: bool=False,
            n_processes: Optional[int]=None,
    ) -> Optional[ImageStack]:
        """Perform filtering of an image stack

        Parameters
        ----------
        stack : ImageStack
            Stack to be filtered.
        in_place : bool
            if True, process ImageStack in-place, otherwise return a new stack
        verbose : bool
            if True, report on filtering progress (default = False)
        n_processes : Optional[int]
            Number of parallel processes to devote to calculating the filter

        Returns
        -------
        Optional[ImageStack] :
            if in-place is False, return the results of filter as a new stack

        """
        low_pass: Callable = partial(self.low_pass, sigma=self.sigma)
        result = stack.apply(
            low_pass, is_volume=self.is_volume, verbose=verbose, in_place=in_place,
            n_processes=n_processes
        )
        if not in_place:
            return result
        return None



import numpy as np
import pandas as pd
import regional

from starfish.spots import TargetAssignment
from starfish.test.codebook.test_metric_decode import intensity_table_factory
from starfish.types import Features, Indices


def bifurcated_regions(n):
    """
    generate an intensity table with diagonal coordinates ((1, 1), (2, 2), ... (n, n)) over 3
    channels and four rounds, where intensities are randomly generated.

    Split the region into two block-diagonal cells which should encompass 1/2 of the total area but
    all of the points in the domain, since intensities is a diagonal table.
    """

    np.random.seed(777)
    data = np.random.random_sample((n, 3, 4))
    diagonal_intensities = intensity_table_factory(data)

    x = diagonal_intensities[Indices.X.value].max() + 1
    y = diagonal_intensities[Indices.Y.value].max() + 1
    box_one_coords = [
        [0, 0],
        [0, np.floor(x / 2)],
        [np.ceil(y / 2), 0],
        [np.floor(y / 2), np.floor(x / 2)]]
    box_two_coords = [
        [np.floor(y / 2), np.floor(x / 2)],
        [np.floor(y / 2), x],
        [y, np.floor(x / 2)],
        [y, x]
    ]
    regions = regional.many([regional.one(box_one_coords), regional.one(box_two_coords)])

    # assign intensity_table some target values that are just sequential numbers
    diagonal_intensities[Features.TARGET] = (Features.AXIS, np.arange(n).astype(str))

    return diagonal_intensities, regions


def test_target_assignment_point_in_poly_2d_all_points_in_cells():
    """Create a set of intensities where all points lie inside the defined cell regions"""

    diagonal_intensities, regions = bifurcated_regions(10)
    ta = TargetAssignment.PointInPoly2D()

    assigned_intensities = ta.run(diagonal_intensities, regions)

    target_counts = pd.Series(
        *np.unique(assigned_intensities[Features.CELL_ID], return_counts=True)[::-1]
    )

    assert target_counts[0] == 5
    assert target_counts[1] == 5


def test_target_assignment_point_in_poly_2d_points_outside_of_cells():
    """Create a set of intensities where no points fall into cells"""

    n = 10
    diagonal_intensities, regions = bifurcated_regions(n)
    diagonal_intensities[Indices.X.value] += 50  # now all x-values lie outside cell regions
    ta = TargetAssignment.PointInPoly2D()

    assigned_intensities = ta.run(diagonal_intensities, regions)

    assert np.array_equal(
        assigned_intensities[Features.CELL_ID].values,
        np.full(n, fill_value=None, dtype=object)
    )



import json
import os
import shutil
import sys
import unittest
from typing import Sequence

import jsonpath_rw
import numpy as np
import pandas as pd

from starfish.intensity_table import IntensityTable
from starfish.types import Features
from starfish.util import exec


def get_jsonpath_from_file(json_filepath_components: Sequence[str], jsonpath: str):
    """
    Given a series of filepath components, join them to find a file <FILE>.  Open that file, and
    locate a specific value in the json structure <PATH>.  Join the directory path of <FILE> and
    <PATH> and return that.

    For example, if json_filepath_components is ["/tmp", "formatted", "experiment.json"] and
    jsonpath is "$['hybridization']", this method will open /tmp/formatted/experiment.json, decode
    that as a json document, and locate the value of the key 'hybridization'.  It will return
    /tmp/formatted/XXX, where XXX is the value of the key.
    """
    json_filepath = os.path.join(*json_filepath_components)
    dirname = os.path.dirname(json_filepath)
    with open(json_filepath, "r") as fh:
        document = json.load(fh)
        return os.path.join(dirname, jsonpath_rw.parse(jsonpath).find(document)[0].value)


class TestWithIssData(unittest.TestCase):
    SUBDIRS = (
        "raw",
        "formatted",
        "registered",
        "filtered",
        "results",
    )

    STAGES = (
        [
            sys.executable,
            "examples/get_iss_data.py",
            lambda tempdir, *args, **kwargs: os.path.join(tempdir, "raw"),
            lambda tempdir, *args, **kwargs: os.path.join(tempdir, "formatted"),
            "--d", "1",
        ],
        [
            "starfish", "registration",
            "--input", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [
                    tempdir,
                    get_jsonpath_from_file(
                        [tempdir, "formatted", "experiment.json"],
                        "$['primary_images']",
                    ),
                ],
                "$['contents']['fov_000']"
            ),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "registered", "hybridization.json"),
            "FourierShiftRegistration",
            "--reference-stack", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [
                    tempdir,
                    get_jsonpath_from_file(
                        [tempdir, "formatted", "experiment.json"],
                        "$['auxiliary_images']['dots']",
                    ),
                ],
                "$['contents']['fov_000']"
            ),
            "--upsampling", "1000",
        ],
        [
            "starfish", "validate", "--experiment-json",
            lambda tempdir: os.sep.join([tempdir, "formatted", "experiment.json"])
        ],
        [
            "starfish", "filter",
            "--input", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "registered", "hybridization.json"),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "hybridization.json"),
            "WhiteTophat",
            "--masking-radius", "15",
        ],
        [
            "starfish", "filter",
            "--input", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [
                    tempdir,
                    get_jsonpath_from_file(
                        [tempdir, "formatted", "experiment.json"],
                        "$['auxiliary_images']['nuclei']",
                    ),
                ],
                "$['contents']['fov_000']"
            ),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "nuclei.json"),
            "WhiteTophat",
            "--masking-radius", "15",
        ],
        [
            "starfish", "filter",
            "--input", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [
                    tempdir,
                    get_jsonpath_from_file(
                        [tempdir, "formatted", "experiment.json"],
                        "$['auxiliary_images']['dots']",
                    ),
                ],
                "$['contents']['fov_000']"
            ),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "dots.json"),
            "WhiteTophat",
            "--masking-radius", "15",
        ],
        [
            "starfish", "detect_spots",
            "--input", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "hybridization.json"),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(tempdir, "results"),
            "GaussianSpotDetector",
            "--blobs-stack", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "dots.json"),
            "--min-sigma", "4",
            "--max-sigma", "6",
            "--num-sigma", "20",
            "--threshold", "0.01",
        ],
        [
            "starfish", "segment",
            "--hybridization-stack", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "hybridization.json"),
            "--nuclei-stack", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "filtered", "nuclei.json"),
            "-o", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "results", "regions.geojson"),
            "Watershed",
            "--dapi-threshold", ".16",
            "--input-threshold", ".22",
            "--min-distance", "57",
        ],
        [
            "starfish", "target_assignment",
            "--coordinates-geojson",
            lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "results", "regions.geojson"),
            "--intensities", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "results", "spots.nc"),
            "--output", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "results", "targeted-spots.nc"),
            "PointInPoly2D",
        ],
        [
            "starfish", "decode",
            "-i", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "results", "targeted-spots.nc"),
            "--codebook", lambda tempdir, *args, **kwargs: get_jsonpath_from_file(
                [tempdir, "formatted", "experiment.json"],
                "$['codebook']",
            ),
            "-o", lambda tempdir, *args, **kwargs: os.path.join(
                tempdir, "results", "decoded-spots.nc"),
            "PerRoundMaxChannelDecoder",
        ],
    )

    def test_run_pipeline(self):

        tempdir = exec.stages(
            TestWithIssData.STAGES,
            TestWithIssData.SUBDIRS,
            keep_data=True,
        )

        try:
            intensities = IntensityTable.load(os.path.join(tempdir, "results", "decoded-spots.nc"))
            genes, counts = np.unique(
                intensities.coords[Features.TARGET], return_counts=True)
            gene_counts = pd.Series(counts, genes)
            assert gene_counts['ACTB_human'] > gene_counts['ACTB_mouse']
        finally:
            if os.getenv("TEST_ISS_KEEP_DATA") is None:
                shutil.rmtree(tempdir)



import regional
from scipy.sparse import coo_matrix


def label_to_regions(labels) -> regional.many:
    label_mat_coo = coo_matrix(labels)

    def region_for(label_mat_coo, label):
        ind = label_mat_coo.data == label
        # TODO does this work in 3D?
        x = label_mat_coo.row[ind]
        y = label_mat_coo.col[ind]

        re = regional.one(list(zip(x, y)))
        return re

    unique_labels = sorted(set(label_mat_coo.data))
    regions = [region_for(label_mat_coo, label) for label in unique_labels]

    return regional.many(regions)



from typing import Tuple, Union

import numpy as np
import pytest

from starfish.image._filter import gaussian_high_pass
from starfish.stack import ImageStack
from starfish.types import Number


def random_data_image_stack_factory():
    data = np.random.uniform(0, 1, 100).reshape(1, 1, 1, 10, 10).astype(np.float32)
    return ImageStack.from_numpy_array(data)


@pytest.mark.parametrize('sigma', (1, (1, 1)))
def test_gaussian_high_pass(sigma: Union[Number, Tuple[Number]]) -> None:
    """high pass is subtractive, sum of array should be less after running."""
    image_stack = random_data_image_stack_factory()
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma)
    sum_before = np.sum(image_stack.numpy_array)
    ghp.run(image_stack)
    assert np.sum(image_stack.numpy_array) < sum_before


@pytest.mark.parametrize('sigma', (1, (1, 1)))
def test_gaussian_high_pass_apply(sigma: Union[Number, Tuple[Number]]) -> None:
    """same as test_gaussian_high_pass, but tests apply loop functionality."""
    image_stack = random_data_image_stack_factory()
    sum_before = np.sum(image_stack.numpy_array)
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma)
    ghp.run(image_stack)
    assert np.sum(image_stack.numpy_array) < sum_before


@pytest.mark.parametrize('sigma', (1, (1, 0, 1)))
def test_gaussian_high_pass_apply_3d(sigma: Union[Number, Tuple[Number]]) -> None:
    """same as test_gaussian_high_pass, but tests apply loop functionality in 3d"""
    image_stack = random_data_image_stack_factory()
    sum_before = np.sum(image_stack.numpy_array)
    ghp = gaussian_high_pass.GaussianHighPass(sigma=sigma, is_volume=True)
    ghp.run(image_stack)
    assert np.sum(image_stack.numpy_array) < sum_before



from typing import Optional

from starfish.pipeline.algorithmbase import AlgorithmBase
from starfish.stack import ImageStack


class FilterAlgorithmBase(AlgorithmBase):
    def run(self, stack: ImageStack) -> Optional[ImageStack]:
        """Performs filtering on an ImageStack."""
        raise NotImplementedError()



import json
from typing import Callable, MutableMapping, MutableSequence, Optional, Sequence, Set, Union

from semantic_version import Version
from slicedimage import Collection, TileSet
from slicedimage.io import Reader, resolve_path_or_url, resolve_url
from slicedimage.urlpath import pathjoin

from starfish.codebook import Codebook
from starfish.stack import ImageStack


class FieldOfView:
    """
    This encapsulates a field of view.  It contains the primary image and auxiliary images that are
    associated with the field of view.

    Auxiliary images can be accessed using a key, i.e., FOV[aux_image_type].

    Properties
    -------
    name                   The name of the FOV.  In an experiment with only a single FOV, this may
                           be None.
    primary_image          The primary image for this field of view.
    auxiliary_image_types  A set of all the auxiliary image types.
    """
    def __init__(
            self,
            name: Optional[str],
            primary_image: Optional[ImageStack]=None,
            auxiliary_images: Optional[MutableMapping[str, ImageStack]]=None,
            primary_image_tileset: Optional[TileSet]=None,
            auxiliary_image_tilesets: Optional[MutableMapping[str, TileSet]]=None,
    ) -> None:
        """
        Fields of views can obtain their primary image from either an ImageStack or a TileSet (but
        only one).  It can obtain their auxiliary image dictionary from either a dictionary of
        auxiliary image name to ImageStack or a dictionary of auxiliary image name to TileSet (but
        only one).

        Note that if the source image is from a TileSet, the decoding of TileSet to ImageStack does
        not happen until the image is accessed.  Be prepared to handle errors when images are
        accessed.
        """
        self._name = name
        self._primary_image: Union[ImageStack, TileSet]
        self._auxiliary_images: MutableMapping[str, Union[ImageStack, TileSet]]
        if primary_image is not None:
            self._primary_image = primary_image
            if primary_image_tileset is not None:
                raise ValueError(
                    "Only one of (primary_image, primary_image_tileset) should be set.")
        elif primary_image_tileset is not None:
            self._primary_image = primary_image_tileset
        else:
            raise ValueError("Field of view must have a primary image")
        if auxiliary_images is not None:
            self._auxiliary_images = auxiliary_images
            if auxiliary_image_tilesets is not None:
                raise ValueError(
                    "Only one of (auxiliary_images, auxiliary_image_tilesets) should be set.")
        elif auxiliary_image_tilesets is not None:
            self._auxiliary_images = auxiliary_image_tilesets
        else:
            self._auxiliary_images = dict()

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def primary_image(self) -> ImageStack:
        if isinstance(self._primary_image, TileSet):
            self._primary_image = ImageStack(self._primary_image)
        return self._primary_image

    @property
    def auxiliary_image_types(self) -> Set[str]:
        return set(self._auxiliary_images.keys())

    def __getitem__(self, item) -> ImageStack:
        if isinstance(self._auxiliary_images[item], TileSet):
            self._auxiliary_images[item] = ImageStack(self._auxiliary_images[item])
        return self._auxiliary_images[item]


class Experiment:
    """
    This encapsulates an experiment, with one or more fields of view and a codebook.  An individual
    FOV can be retrieved using a key, i.e., experiment[fov_name].

    Constructors
    -------
    from_json     Given a URL or a path to an experiment.json document, return an Experiment object
                  corresponding to the document.

    Methods
    -------
    fov           Given a callable that accepts a FOV, return the first FOVs that the callable
                  returns True when passed the FOV.  Because there is no guaranteed sorting for the
                  FOVs, use this cautiously.
    fovs          Given a callable that accepts a FOV, return all the FOVs that the callable returns
                  True when passed the FOV.
    fovs_by_name  Given one or more FOV names, return the FOVs that match those names.

    Properties
    -------
    codebook      Returns the codebook associated with this experiment.
    extras        Returns the extras dictionary associated with this experiment.
    """
    CURRENT_VERSION = Version("3.0.0")
    MIN_SUPPORTED_VERSION = Version("3.0.0")
    MAX_SUPPORTED_VERSION = Version("3.0.0")

    def __init__(
            self,
            fovs: Sequence[FieldOfView],
            codebook: Codebook,
            extras: dict,
            *,
            src_doc: dict=None,
    ) -> None:
        self._fovs = fovs
        self._codebook = codebook
        self._extras = extras
        self._src_doc = src_doc

    @classmethod
    def from_json(cls, json_url: str) -> "Experiment":
        """
        Construct an `Experiment` from an experiment.json file format specifier

        Parameters
        ----------
        json_url : str
            file path or web link to an experiment.json file

        Returns
        -------
        Experiment :
            Experiment object serving the requested experiment data

        """
        backend, name, baseurl = resolve_path_or_url(json_url)
        with backend.read_contextmanager(name) as fh:
            experiment_document = json.load(fh)

        cls.verify_version(experiment_document['version'])

        _, codebook_name, codebook_baseurl = resolve_url(experiment_document['codebook'], baseurl)
        codebook_absolute_url = pathjoin(codebook_baseurl, codebook_name)
        codebook = Codebook.from_json(codebook_absolute_url)

        extras = experiment_document['extras']

        primary_image: Collection = Reader.parse_doc(experiment_document['primary_images'], baseurl)
        auxiliary_images: MutableMapping[str, Collection] = dict()
        for aux_image_type, aux_image_url in experiment_document['auxiliary_images'].items():
            auxiliary_images[aux_image_type] = Reader.parse_doc(aux_image_url, baseurl)

        fovs: MutableSequence[FieldOfView] = list()
        for fov_name, primary_tileset in primary_image.all_tilesets():
            aux_image_tilesets_for_fov: MutableMapping[str, TileSet] = dict()
            for aux_image_type, aux_image_collection in auxiliary_images.items():
                aux_image_tileset = aux_image_collection.find_tileset(fov_name)
                if aux_image_tileset is not None:
                    aux_image_tilesets_for_fov[aux_image_type] = aux_image_tileset

            fov = FieldOfView(
                fov_name,
                primary_image_tileset=primary_tileset,
                auxiliary_image_tilesets=aux_image_tilesets_for_fov,
            )
            fovs.append(fov)

        return Experiment(fovs, codebook, extras, src_doc=experiment_document)

    @classmethod
    def verify_version(cls, semantic_version_str: str) -> None:
        version = Version(semantic_version_str)
        if not (Experiment.MIN_SUPPORTED_VERSION <= version <= Experiment.MAX_SUPPORTED_VERSION):
            raise ValueError(
                f"version {version} not supported.  This version of the starfish library only "
                f"supports formats from {Experiment.MIN_SUPPORTED_VERSION} to "
                f"{Experiment.MAX_SUPPORTED_VERSION}")

    def fov(
            self,
            filter_fn: Callable[[FieldOfView], bool]=lambda _: True,
    ) -> FieldOfView:
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.  Return the first
        FOV such that filter_fn(FOV) returns True.  Because there is no guaranteed order for the
        FOVs, use this cautiously.

        If no FOV matches, raise LookupError.
        """
        for fov in self._fovs:
            if filter_fn(fov):
                return fov
        raise LookupError("Cannot find any FOV that the filter allows.")

    def fovs(
            self,
            filter_fn: Callable[[FieldOfView], bool]=lambda _: True,
    ) -> Sequence[FieldOfView]:
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.  Return a list of
        FOVs such that filter_fn(FOV) returns True.
        """
        results: MutableSequence[FieldOfView] = list()
        for fov in self._fovs:
            if not filter_fn(fov):
                continue

            results.append(fov)
        return results

    def fovs_by_name(self, *names):
        """
        Given a callable filter_fn, apply it to all the FOVs in this experiment.  Return a list of
        FOVs such that filter_fn(FOV) returns True.
        """
        return self.fovs(filter_fn=lambda fov: fov.name in names)

    def __getitem__(self, item):
        fovs = self.fovs_by_name(item)
        if len(fovs) == 0:
            raise IndexError(f"No field of view with name \"{item}\"")
        return fovs[0]

    @property
    def codebook(self) -> Codebook:
        return self._codebook

    @property
    def extras(self):
        return self._extras



import pytest
from pkg_resources import resource_filename

from .util import SpaceTxValidator

schema_path = resource_filename("validate_sptx", "schema/fov_manifest.json")
validator = SpaceTxValidator(schema_path)


def test_fov_manifest():
    fov_manifest_example_path = resource_filename(
        "validate_sptx", "examples/fov_manifest/fov_manifest.json")
    assert validator.validate_file(fov_manifest_example_path)


def test_empty_manifest_raises_validation_error():
    with pytest.warns(UserWarning):
        assert not validator.validate_object({})



import argparse
import json
from typing import MutableMapping

from starfish.types import Indices
from starfish.util.argparse import FsExistsType
from . import AUX_IMAGE_NAMES, write_experiment_json


class StarfishIndex:
    def __call__(self, spec_json):
        try:
            spec = json.loads(spec_json)
        except json.decoder.JSONDecodeError:
            raise argparse.ArgumentTypeError(
                "Could not parse {} into a valid index specification.".format(spec_json))

        return {
            Indices.ROUND: spec.get(Indices.ROUND, 1),
            Indices.CH: spec.get(Indices.CH, 1),
            Indices.Z: spec.get(Indices.Z, 1),
        }


class Cli:
    parser_group = None

    """
    Maps the name of an aux image to the name in the parsed arguments object.
    """
    name_arg_map: MutableMapping[str, str] = dict()

    @staticmethod
    def add_to_parser(parser):
        parser.add_argument(
            "output_dir",
            type=FsExistsType())
        parser.add_argument(
            "--fov-count",
            type=int,
            required=True,
            help="Number of FOVs in this experiment.")
        parser.add_argument(
            "--hybridization-dimensions",
            type=StarfishIndex(),
            required=True,
            help="Dimensions for the hybridization images.  Should be a json dict, with {}, {}, "
                 "and {} as the possible keys.  The value should be the shape along that "
                 "dimension.  If a key is not present, the value is assumed to be 0.".format(
                Indices.ROUND.value,
                Indices.CH.value,
                Indices.Z.value))
        for aux_image_name in AUX_IMAGE_NAMES:
            arg = parser.add_argument(
                "--{}-dimensions".format(aux_image_name),
                type=StarfishIndex(),
                help="Dimensions for the {} images.  Should be a json dict, with {}, {}, and {} as "
                     "the possible keys.  The value should be the shape along that dimension.  If "
                     "a key is not present, the value is assumed to be 0.".format(
                    aux_image_name, Indices.ROUND, Indices.CH, Indices.Z))
            Cli.name_arg_map[aux_image_name] = arg.dest
        parser.set_defaults(starfish_command=Cli.run)

        Cli.parser_group = parser

    @staticmethod
    def run(args, print_help=False):
        write_experiment_json(
            args.output_dir, args.fov_count, args.hybridization_dimensions,
            {
                aux_image_name: getattr(args, Cli.name_arg_map[aux_image_name])
                for aux_image_name in AUX_IMAGE_NAMES
            }
        )
